
#pragma once
//#include "../myincludes.hh"
#include <cmath>

template<typename GV, typename RF, std::size_t dim_range>
class ZeroVectorFunction :
  public Dune::PDELab::AnalyticGridFunctionBase<
  Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim_range>,
  ZeroVectorFunction<GV,RF,dim_range> >,
  public Dune::PDELab::InstationaryFunctionDefaults
{
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim_range> Traits;
  typedef Dune::PDELab::AnalyticGridFunctionBase<Traits, ZeroVectorFunction> BaseT;

  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  ZeroVectorFunction(const GV & gv) : BaseT(gv) {}

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y=0;
  }
};

template<typename GV, typename RF>
class ZeroScalarFunction
  : public ZeroVectorFunction<GV,RF,1>
{
public:

  ZeroScalarFunction(const GV & gv) : ZeroVectorFunction<GV,RF,1>(gv) {}

};



template<typename GV, typename RF, int dim>
class PhiZero :
  public Dune::PDELab::AnalyticGridFunctionBase<Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim>, PhiZero<GV,RF,dim> >
{
  const RF mean;
  const RF maxnoise;
  const RF eps;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim> Traits;
  typedef Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiZero<GV,RF,dim> > BaseT;
  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  PhiZero(const GV & gv, const RF& mean_, const RF& maxnoise_, const RF& eps_) :
     BaseT(gv), mean(mean_), maxnoise(maxnoise_), eps(eps_) {}

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y = 0.0;
    //Random Noise
    //y[0] = mean+ maxnoise*(-1 + 2*((RF)rand())/RAND_MAX);
    //Deterministic "pseudeorandom"
    //y[0] = mean+ maxnoise* sin(sqrt(3)*100000*x[0]+sqrt(2))*cos(sqrt(2)*1000000*x[1]+sqrt(5))

    //Bubble
    //RF radius = sqrt( (x[0]-0.5)*(x[0]-0.5) + (x[1]-0.5)*(x[1]-0.5) );
    //y[0] = 0.5*(1+tanh(1/eps/sqrt(2)*(radius-0.2)));

    //Stripes
    // RF d = 0;
    // if(x[0]<3./8)
    //   d = x[0]-1./4;
    // else if(x[0] < 5./8)
    //   d = 1./8 - (x[0]-3./8);
    // else
    //   d = -1./8 + (x[0]-5./8);
    // y[0] = 0.5*(1+tanh(1/eps/sqrt(2)*d));

    //Left/Right
    //y[0] = 0.5*(1+tanh(1/eps/sqrt(2)*(x[0]-0.75)));

    //Top/Bottom (Rayleigh-Taylor Instability)
    y[0] = 0.5*(1-tanh(1/eps/sqrt(2)*(x[1]-0.75-0.01*cos(x[0]*2*3.1415))));
  }
};

template<typename GV, typename RF, int dim>
class ThreephasePhiZero :
  public Dune::PDELab::AnalyticGridFunctionBase<Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim>, ThreephasePhiZero<GV,RF,dim> >
{
  const RF mean;
  const RF maxnoise;
  const RF eps;
  const int phaseNumber;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim> Traits;
  typedef Dune::PDELab::AnalyticGridFunctionBase<Traits,ThreephasePhiZero<GV,RF,dim> > BaseT;
  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  ThreephasePhiZero(const GV & gv, const int& num_, const RF& mean_, const RF& maxnoise_, const RF& eps_) :
     BaseT(gv), mean(mean_), maxnoise(maxnoise_), eps(eps_), phaseNumber(num_) {}

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y = 0.0;

    //Stripes
    // if(phaseNumber == 0)
    // {
    //   RF d = (x[0]-3./4);
    //   y[0] = 0.5*(1+tanh(1/eps/sqrt(2)*d));
    // }
    // if(phaseNumber == 1)
    // {
    //   RF d = -(x[0]-1./4);
    //   y[0] = 0.5*(1+tanh(1/eps/sqrt(2)*d));
    // }

    //Contact point
    RF d1 = x[0]-0.5;
    RF yx = 0.5*(1+tanh(1/eps/sqrt(2)*d1));

    RF d2 = x[1]-0.75;
    if(phaseNumber == 0)
      d2 = -d2;
    y[0]=0.5*(1+tanh(1/eps/sqrt(2)*d2))*yx;

  }
};

template<typename GV, typename RF, int dim>
class UZero :
  public Dune::PDELab::AnalyticGridFunctionBase<Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim>, UZero<GV,RF,dim> >
{
  const RF mean;
  const RF maxnoise;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim> Traits;
  typedef Dune::PDELab::AnalyticGridFunctionBase<Traits,UZero<GV,RF,dim> > BaseT;
  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  UZero(const GV & gv, const RF& mean_, const RF& maxnoise_) :
     BaseT(gv), mean(mean_), maxnoise(maxnoise_) {}

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y[0] = mean;
    //if (x[0]> 1-1e-5)
    //{
    //  y[0]=mean-0.3;
    //}
  }
};



class U_BC : public Dune::PDELab::DirichletConstraintsParameters
{
public:
  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if (xg[0]> 1-1e-5)
    {
      return true;
    }
    return false;
  }
};


class SinglePointPressureBC : public Dune::PDELab::DirichletConstraintsParameters
{
public:
  template<typename I> bool isDirichlet(const I & intersection,
                                        const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    //std::cout<< "x0: " << xg[0] << "    x1: " << xg[1] << std::endl;
    if (xg[0]< 1e-2)
    {
      //std::cout<< "Found Pressure Zero Point";
      //return true;
    }
    return false;
  }
};

template<typename GV, typename RF, int dim>
class IniVelocity :
  public Dune::PDELab::AnalyticGridFunctionBase<
  Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim>,
  IniVelocity<GV,RF,dim> >
{
  const RF meanflow_;
  const RF timemax;
  RF time;


public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim> Traits;
  typedef Dune::PDELab::AnalyticGridFunctionBase<Traits,IniVelocity<GV,RF,dim> > BaseT;

  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  IniVelocity(const GV & gv, const RF& meanflow, const RF& tm) :
    BaseT(gv)
    , meanflow_(meanflow), timemax(tm)
  { time=0.0;}

  template <typename T>
  void setTime(T t){
    time = t;
  }

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y = 0.0;
    RF radius = sqrt( (x[0]-0.5)*(x[0]-0.5) + (x[1]-0.5)*(x[1]-0.5) );
    RF r = meanflow_*(0.5-radius);
    if(r< 0)
    {
      r=0;
    }

    //y[0] = -r*(x[1]-0.5) * 1e-4;     //Rotating flow
    //y[1] = r*(x[0]-0.5)  * 1e-4;

    //y[1]=x[0];              //Shear flow
    //y[0]=0;

    //y[0]=0;                   //No flow
    //y[1]=0;
    RF tscale=fmin(time/timemax,1);

    y[0]=0;                   //Shear flow boundary
    if(x[0]>0.9999)
    {
      //y[1]=4*meanflow_*x[1]*(1-x[1])*tscale;
      if (x[1]>0 && x[1]<0.75)
      {
        y[1]=64/9*meanflow_*x[1]*(0.75-x[1])*tscale;
      }
    }
  }
};


template<typename GV, typename RF, int dim>
class TUPressure :
  public Dune::PDELab::AnalyticGridFunctionBase<
  Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim>,
  TUPressure<GV,RF,dim> >
{
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim> Traits;
  typedef Dune::PDELab::AnalyticGridFunctionBase<Traits,TUPressure<GV,RF,dim> > BaseT;

  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  TUPressure(const GV & gv) :
    BaseT(gv)
  { }

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y = 0.0;
    //y[0]=0.1*x[0];
  }
  template <typename T>  void setTime(T t){}
};
